///////////////////////////////////////////////////////////////////////////////////////////////////
// Company: <Name>
//
// File: WideMultTop.v
// File history:
//      <Revision number>: <Date>: <Comments>
//      <Revision number>: <Date>: <Comments>
//      <Revision number>: <Date>: <Comments>
//
// The calculation is dout= A + B +(C *D) +(C*C*E), A + B + C(D + (C*E))
// B, C is 64bit width, A,D,E is 128 bit width. 
// A,C,D,E  are signed signal.
// For the A,D,E that are 128 bit width, it all include the 64 fractional bits, 63 integer bits and  1 sign bit.
// For B and C,  it doesnt include the fractional bit. And B is unsigned, C is a signed signal.
// The output of module is 261bits, but after remove the fractional and unused MSBs, 
// actually we only need the [128:65]bit of it ,and its an unsigned signal.

// The module instantiates multiplier blocks using an inferred 64x64 bit multiplier as the base building block.
// The 64x64 bit multiplier is pipelined at each multiply stage for max performance.
//
// <Description here>
//
// Targeted device: <Family::PolarFire> <Die::MPF100T> <Package::FCG484>
// Author: <Name>
//
/////////////////////////////////////////////////////////////////////////////////////////////////// 

////////////////////////////////////////////////////////////////////////////////////////
//DataFlow
//
//
//	----------	  ------ 	      -----						    -----
//	| C X E  | -> | /2 | -> + --> |-1 | -->  X -->  + --> + --> |-1 |
//	----------	  ------    |	  -----	     |      |     |	    -----
//	    ------	            |			     |      |     |
//  D-> |-17 |---------------			     |      |     |  
//	    ------					             |      |     |								   
//	    ------					             |      |     |
//  C-> |-18 |--------------------------------      |     |
//	    ------                                      |     |
//	    ------                                      |     |
//  B->	|-36 |---------------------------------------     |
//	    ------                                            |
//	    ------                                            |
//  A->	|-36 |---------------------------------------------
//	    ------
////////////////////////////////////////////////////////////////////////////////////////

//`timescale <time_units> / <precision>


module WideMultTop( resetn, clock ,A, B, C, D, E, WE, result );
input           resetn;
input           clock;
input  signed  [127:0]  A;
input          [63:0]   B;
input  signed  [63:0]   C;
input  signed  [127:0]  D;
input  signed  [127:0]  E;
input                   WE;             // Assert high to start operations
output reg signed  [255:0] result;      // result

reg signed [127:0]  d_del;              // 17 clock cycle delay
reg signed [63:0]   c_del;              // 18 clock cycle delay
reg signed [127:0]  a_del;              // 36 clock cycle delay
reg signed [63:0]   b_del;              // 36 clock cycle delay
reg signed [191:0]  ce_result;          // C*E
reg signed [191:0]  cepd_result;        // (C*E) +D 
reg signed [255:0]  ccepd_result;       // (C*C*E) +D
reg signed [191:0]  ce_resultrs;        // (C * E) /2

// instantiate a 128x64 bit multiplier which is built from 
// a 64x65 bit multiplier and a 64x64 bit multiplier
// used to calculate C*E
mult128x64 mult128x64_ce( 
    .resetn     (resetn ), 
    .clock      (clock ), 
    .A          (C ), 
    .B          (E ), 
    .prodab     (ce_result )
);

// right shift output of 128x64 multiply to divide output by 2.
assign ce_resultrs = {ce_result[191], ce_result[191:1]};

// instantiate a 192x64 bit multiplier which is built from 
// two 64x65 bit multipliers and a 64x64 bit multiplier
// used to calculate C*C*E 
mult192x64 mult192x64_inst( 
    .resetn     (resetn ), 
    .clock      (clock ), 
    .A          (c_del), 
    .B          (cepd_result ), 
    .prodab_del (ccepd_result )
);

// delay for D input to line up with output of C * E
DelayFIFOx128_18 DelayFIFOx128_18_inst( 
    .resetn     (resetn ), 
    .clock      (clock  ), 
    .WE         (WE ), 
    .datain     (D ), 
    .dataout    (d_del ) 
);

// delay for D input to line up with registered output of 0.5(C * E) + D
DelayFIFOx64_19 DelayFIFOx64_19_inst( 
    .resetn     (resetn ), 
    .clock      (clock  ), 
    .WE         (WE ), 
    .datain     (C ), 
    .dataout    (c_del ) 
);

// delay for A input to line up with registered output of (0.5(C * E) + D)C
DelayFIFOx128_38 DelayFIFOx128_38_inst( 
    .resetn     (resetn ), 
    .clock      (clock  ), 
    .WE         (WE ), 
    .datain     (A ), 
    .dataout    (a_del ) 
);

// delay for B input to line up with registered output of (0.5(C * E) + D)C
DelayFIFOx64_38 DelayFIFOx64_38_inst( 
    .resetn     (resetn ), 
    .clock      (clock  ), 
    .WE         (WE ), 
    .datain     (B ), 
    .dataout    (b_del ) 
);

// register results to improve performance
always @ (negedge resetn or posedge clock) begin
    if (!resetn) begin
        cepd_result  <= 192'b0;
        result <= 256'b0;
    end
    else begin
        cepd_result  <= ce_resultrs + d_del;
        result <= a_del + {b_del,{64{1'b0}}} + ccepd_result;
    end
end

endmodule

